# === sendOrderMsg_Media ===
from pymammotion.proto import luba_msg_pb2, luba_mul_pb2
from pymammotion.proto.luba_mul import MUL_LANGUAGE


class MessageMedia:
    @staticmethod
    def send_order_msg_media(self, mul):
        luba_msg = luba_msg_pb2.LubaMsg(
            msgtype=luba_msg_pb2.MSG_CMD_TYPE_MUL,
            sender=luba_msg_pb2.DEV_MOBILEAPP,
            rcver=luba_msg_pb2.SOC_MODULE_MULTIMEDIA,
            msgattr=luba_msg_pb2.MSG_ATTR_REQ,
            seqs=1,
            version=1,
            subtype=1,
            mul=mul,
        )

        return luba_msg.SerializeToString()

    def set_car_volume(self, volume: int):
        return self.send_order_msg_media(luba_mul_pb2.SocMul(set_audio=luba_mul_pb2.MulSetAudio(at_switch=volume)))

    def set_car_voice_language(self, language_type: MUL_LANGUAGE | str | None):
        return self.send_order_msg_media(
            luba_mul_pb2.SocMul(set_audio=luba_mul_pb2.MulSetAudio(au_language=language_type))
        )

    def set_car_wiper(self, round_num: int):
        return self.send_order_msg_media(luba_mul_pb2.SocMul(set_wiper=luba_mul_pb2.MulSetWiper(round=round_num)))



from dataclasses import dataclass, field

from mashumaro.mixins.orjson import DataClassORJSONMixin

from pymammotion.utility.device_type import DeviceType


@dataclass
class DeviceLimits(DataClassORJSONMixin):
    blade_height_min: int = 30
    blade_height_max: int = 70
    working_speed_min: float = 0.2
    working_speed_max: float = 1.2
    working_path_min: int = 15
    working_path_max: int = 35


@dataclass
class OperationSettings(DataClassORJSONMixin):
    """Operation settings for a device."""

    is_mow: bool = True
    is_dump: bool = True
    is_edge: bool = False
    collect_grass_frequency: int = 10
    job_mode: int = 0  # taskMode
    job_version: int = 0
    job_id: int = 0
    speed: float = 0.3
    ultra_wave: int = 2  # touch no touch etc
    channel_mode: int = 0  # line mode is grid single double or single2
    channel_width: int = 25
    rain_tactics: int = 0
    blade_height: int = 0
    path_order: str = ""
    toward: int = 0  # is just angle
    toward_included_angle: int = 0
    toward_mode: int = 0  # angle type relative etc
    border_mode: int = 1  # border laps
    obstacle_laps: int = 1
    start_progress: int = 0
    areas: list[int] = field(default_factory=list)


def create_path_order(operation_mode: OperationSettings, device_name: str) -> str:
    i = 8
    bArr = bytearray(8)
    bArr[0] = operation_mode.border_mode
    bArr[1] = operation_mode.obstacle_laps
    bArr[3] = int(operation_mode.start_progress)
    bArr[2] = 0

    if not DeviceType.is_luba1(device_name):
        bArr[4] = 0
        if DeviceType.is_yuka(device_name):
            i = calculate_yuka_mode(operation_mode)
        elif not DeviceType.is_luba_2(device_name):
            i = 0
        bArr[5] = i
        if operation_mode.is_dump:
            b = int(operation_mode.collect_grass_frequency)
        else:
            b = 10
        bArr[6] = b
    if DeviceType.is_luba1(device_name):
        bArr[4] = operation_mode.toward_mode
    return bArr.decode()


def calculate_yuka_mode(operation_mode: OperationSettings) -> int:
    if operation_mode.is_mow and operation_mode.is_dump and operation_mode.is_edge:
        return 14
    if operation_mode.is_mow and operation_mode.is_dump and not operation_mode.is_edge:
        return 12
    if operation_mode.is_mow and not operation_mode.is_dump and operation_mode.is_edge:
        return 10
    if operation_mode.is_mow and not operation_mode.is_dump and not operation_mode.is_edge:
        return 8
    if not operation_mode.is_mow and operation_mode.is_dump and operation_mode.is_edge:
        return 6
    if not operation_mode.is_mow and not operation_mode.is_dump and operation_mode.is_edge:
        return 2
    if not operation_mode.is_mow and operation_mode.is_dump and not operation_mode.is_edge:
        return 4
    return 0



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: pymammotion/proto/mctrl_driver.proto
# plugin: python-betterproto
from dataclasses import dataclass


import betterproto


@dataclass
class DrvMotionCtrl(betterproto.Message):
    set_linear_speed: int = betterproto.int32_field(1)
    set_angular_speed: int = betterproto.int32_field(2)


@dataclass
class DrvKnifeHeight(betterproto.Message):
    knife_height: int = betterproto.int32_field(1)


@dataclass
class DrvSrSpeed(betterproto.Message):
    rw: int = betterproto.int32_field(1)
    speed: float = betterproto.float_field(2)


@dataclass
class DrvKnifeStatus(betterproto.Message):
    knife_status: int = betterproto.int32_field(1)


@dataclass
class DrvKnifeChangeReport(betterproto.Message):
    is_start: int = betterproto.int32_field(1)
    start_high: int = betterproto.int32_field(2)
    end_high: int = betterproto.int32_field(3)


@dataclass
class DrvMowCtrlByHand(betterproto.Message):
    main_ctrl: int = betterproto.int32_field(1)
    cut_knife_ctrl: int = betterproto.int32_field(2)
    cut_knife_height: int = betterproto.int32_field(3)
    max_run__speed: float = betterproto.float_field(4)


@dataclass
class RtkCfgReqT(betterproto.Message):
    cmd_length: int = betterproto.int32_field(1)
    cmd_req: str = betterproto.string_field(2)


@dataclass
class RtkCfgReqAckT(betterproto.Message):
    cmd_length: int = betterproto.int32_field(1)
    cmd_response: str = betterproto.string_field(2)


@dataclass
class RtkSysMaskQueryT(betterproto.Message):
    sat_system: int = betterproto.uint32_field(1)


@dataclass
class RtkSysMaskQueryAckT(betterproto.Message):
    sat_system: int = betterproto.uint32_field(1)
    system_mask_bits: list[int] = betterproto.uint32_field(2)


@dataclass
class MctlDriver(betterproto.Message):
    todev_devmotion_ctrl: "DrvMotionCtrl" = betterproto.message_field(
        1, group="SubDrvMsg"
    )
    todev_knife_height_set: "DrvKnifeHeight" = betterproto.message_field(
        2, group="SubDrvMsg"
    )
    bidire_speed_read_set: "DrvSrSpeed" = betterproto.message_field(
        3, group="SubDrvMsg"
    )
    bidire_knife_height_report: "DrvKnifeHeight" = betterproto.message_field(
        4, group="SubDrvMsg"
    )
    toapp_knife_status: "DrvKnifeStatus" = betterproto.message_field(
        5, group="SubDrvMsg"
    )
    mow_ctrl_by_hand: "DrvMowCtrlByHand" = betterproto.message_field(
        6, group="SubDrvMsg"
    )
    rtk_cfg_req: "RtkCfgReqT" = betterproto.message_field(7, group="SubDrvMsg")
    rtk_cfg_req_ack: "RtkCfgReqAckT" = betterproto.message_field(8, group="SubDrvMsg")
    rtk_sys_mask_query: "RtkSysMaskQueryT" = betterproto.message_field(
        9, group="SubDrvMsg"
    )
    rtk_sys_mask_query_ack: "RtkSysMaskQueryAckT" = betterproto.message_field(
        10, group="SubDrvMsg"
    )
    toapp_knife_status_change: "DrvKnifeChangeReport" = betterproto.message_field(
        11, group="SubDrvMsg"
    )



import base64


class DatatypeConverter:
    encode_map = None

    @staticmethod
    def init_encode_map():
        """Initialize the encode map for DatatypeConverter if it is not already
        initialized.

        This function initializes the encode map for DatatypeConverter by
        creating a list of 64 elements and populating it with characters for
        encoding. If the encode map is already initialized, it returns the
        existing encode map.

        Returns:
            list: The encode map for DatatypeConverter.

        """

        if DatatypeConverter.encode_map is None:
            cArr: list[str | int] = [0] * 64
            for num in range(26):
                cArr[num] = chr(num + 65)
            for num_2 in range(26, 52):
                cArr[num_2] = chr(num_2 - 26 + 97)
            for num_3 in range(52, 62):
                cArr[num_3] = chr(num_3 - 52 + 48)
            cArr[62] = "+"
            cArr[63] = "/"
            DatatypeConverter.encode_map = cArr
        return DatatypeConverter.encode_map

    @staticmethod
    def parseBase64Binary(s):
        return base64.b64decode(s)

    @staticmethod
    def printBase64Binary(bArr):
        return DatatypeConverter._printBase64Binary(bArr)

    @staticmethod
    def encode(i):
        return DatatypeConverter.encode_map[i & 63]

    @staticmethod
    def _printBase64Binary(bArr: bytes, i: int = 0, i2=None):
        """Print the Base64 binary representation of a byte array.

        This function takes a byte array and optional start and end indices to
        print the Base64 binary representation.

        Args:
            bArr (list): A list of bytes to be converted to Base64 binary.
            i (int): The starting index of the byte array (default is 0).
            i2 (int): The ending index of the byte array (default is the length of bArr).

        Returns:
            str: The Base64 binary representation of the input byte array.

        """

        if i2 is None:
            i2 = len(bArr)
        cArr = [""] * (((i2 + 2) // 3) * 4)
        DatatypeConverter._printBase64Binary_core(bArr, i, i2, cArr, 0)
        return "".join(cArr)

    @staticmethod
    def _printBase64Binary_core(bArr: bytes, i, i2, cArr, i3):
        """Encode binary data into Base64 format.

        This function encodes binary data into Base64 format following the
        Base64 encoding algorithm.

        Args:
            bArr (list): List of binary data to be encoded.
            i (int): Starting index of the binary data to be encoded.
            i2 (int): Length of binary data to be encoded.
            cArr (list): List to store the encoded Base64 characters.
            i3 (int): Starting index in the cArr to store the encoded characters.

        Returns:
            int: The index in cArr where encoding ends.

        Note:
            This function assumes that DatatypeConverter has a method 'encode' and
            'init_encode_map' for encoding.

        """

        DatatypeConverter.init_encode_map()  # Ensure encode_map is initialized
        while i2 >= 3:
            cArr[i3] = DatatypeConverter.encode(bArr[i] >> 2)
            cArr[i3 + 1] = DatatypeConverter.encode(((bArr[i] & 3) << 4) | ((bArr[i + 1] >> 4) & 15))
            cArr[i3 + 2] = DatatypeConverter.encode(((bArr[i + 1] & 15) << 2) | ((bArr[i + 2] >> 6) & 3))
            cArr[i3 + 3] = DatatypeConverter.encode(bArr[i + 2] & 63)
            i2 -= 3
            i += 3
            i3 += 4

        if i2 == 1:
            cArr[i3] = DatatypeConverter.encode(bArr[i] >> 2)
            cArr[i3 + 1] = DatatypeConverter.encode((bArr[i] & 3) << 4)
            cArr[i3 + 2] = "="
            cArr[i3 + 3] = "="

        if i2 == 2:
            cArr[i3] = DatatypeConverter.encode(bArr[i] >> 2)
            cArr[i3 + 1] = DatatypeConverter.encode(((bArr[i] & 3) << 4) | ((bArr[i + 1] >> 4) & 15))
            cArr[i3 + 2] = DatatypeConverter.encode((bArr[i + 1] & 15) << 2)
            cArr[i3 + 3] = "="

        return i3


# Usage Example:
# converter = DatatypeConverter()
# encoded = converter.printBase64Binary(b"Hello, World!")
# print(encoded)  # Output: "SGVsbG8sIFdvcmxkIQ=="
# decoded = converter.parseBase64Binary(encoded)
# print(decoded)  # Output: b'Hello, World!'



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: pymammotion/proto/mctrl_pept.proto
# plugin: python-betterproto
from dataclasses import dataclass


import betterproto


@dataclass
class PerceptionObstaclesT(betterproto.Message):
    label: int = betterproto.int32_field(1)
    num: int = betterproto.int32_field(2)
    points_x: list[int] = betterproto.int32_field(3)
    points_y: list[int] = betterproto.int32_field(4)


@dataclass
class PerceptionObstaclesVisualizationT(betterproto.Message):
    is_heart_beat: int = betterproto.int32_field(1)
    num: int = betterproto.int32_field(2)
    obstacles: list["PerceptionObstaclesT"] = betterproto.message_field(3)
    timestamp: float = betterproto.double_field(4)
    scale: float = betterproto.float_field(5)


@dataclass
class PerceptionUniversalBuffT(betterproto.Message):
    perception_type: int = betterproto.int32_field(1)
    perception_len: int = betterproto.int32_field(2)
    universal_buff: list[int] = betterproto.int64_field(3)


@dataclass
class MctlPept(betterproto.Message):
    perception_obstacles_visualization: "PerceptionObstaclesVisualizationT" = (
        betterproto.message_field(1, group="SubPeptMsg")
    )
    perception_universal_buff: "PerceptionUniversalBuffT" = betterproto.message_field(
        2, group="SubPeptMsg"
    )



import asyncio
import base64
import json
import logging
from collections import deque
from typing import Any, Awaitable, Callable, Optional, cast

import betterproto

from pymammotion import CloudIOTGateway, MammotionMQTT
from pymammotion.aliyun.dataclass.dev_by_account_response import Device
from pymammotion.data.model.device import MowingDevice
from pymammotion.data.mqtt.event import ThingEventMessage
from pymammotion.event.event import DataEvent
from pymammotion.mammotion.commands.mammotion_command import MammotionCommand
from pymammotion.mammotion.devices.base import MammotionBaseDevice
from pymammotion.mqtt.mammotion_future import MammotionFuture
from pymammotion.proto import has_field
from pymammotion.proto.luba_msg import LubaMsg

_LOGGER = logging.getLogger(__name__)


class MammotionCloud:
    """Per account MQTT cloud."""

    def __init__(self, mqtt_client: MammotionMQTT, cloud_client: CloudIOTGateway) -> None:
        self.cloud_client = cloud_client
        self.loop = asyncio.get_event_loop()
        self._ble_sync_task = None
        self.is_ready = False
        self.command_queue = asyncio.Queue()
        self._waiting_queue = deque()
        self.mqtt_message_event = DataEvent()
        self.on_ready_event = DataEvent()
        self._operation_lock = asyncio.Lock()
        self._mqtt_client = mqtt_client
        self._mqtt_client.on_connected = self.on_connected
        self._mqtt_client.on_disconnected = self.on_disconnected
        self._mqtt_client.on_message = self._on_mqtt_message
        self._mqtt_client.on_ready = self.on_ready

        # temporary for testing only
        # self._start_sync_task = self.loop.call_later(30, lambda: asyncio.ensure_future(self.start_sync(0)))

    async def on_ready(self) -> None:
        loop = asyncio.get_event_loop()
        loop.create_task(self.process_queue())
        await self.on_ready_event.data_event(None)

    def is_connected(self) -> bool:
        return self._mqtt_client.is_connected

    def disconnect(self) -> None:
        self._mqtt_client.disconnect()

    def connect_async(self) -> None:
        self._mqtt_client.connect_async()

    def send_command(self, iot_id: str, command: bytes) -> None:
        self._mqtt_client.get_cloud_client().send_cloud_command(iot_id, command)

    async def on_connected(self) -> None:
        """Callback for when MQTT connects."""

    async def on_disconnected(self) -> None:
        """Callback for when MQTT disconnects."""

    async def process_queue(self) -> None:
        while True:
            # Get the next item from the queue
            iot_id, key, command, future = await self.command_queue.get()
            try:
                # Process the command using _execute_command_locked
                result = await self._execute_command_locked(iot_id, key, command)
                # Set the result on the future
                future.set_result(result)
            except Exception as ex:
                # Set the exception on the future if something goes wrong
                future.set_exception(ex)
            finally:
                # Mark the task as done
                self.command_queue.task_done()

    async def _execute_command_locked(self, iot_id: str, key: str, command: bytes) -> bytes:
        """Execute command and read response."""
        assert self._mqtt_client is not None
        self._key = key
        _LOGGER.debug("Sending command: %s", key)

        await self.loop.run_in_executor(None, self._mqtt_client.get_cloud_client().send_cloud_command, iot_id, command)
        future = MammotionFuture(iot_id)
        self._waiting_queue.append(future)
        timeout = 5
        try:
            notify_msg = await future.async_get(timeout)
        except asyncio.TimeoutError:
            notify_msg = b""

        _LOGGER.debug("%s: Message received", iot_id)

        return notify_msg

    async def _on_mqtt_message(self, topic: str, payload: str, iot_id: str) -> None:
        """Handle incoming MQTT messages."""
        _LOGGER.debug("MQTT message received on topic %s: %s, iot_id: %s", topic, payload, iot_id)

        json_str = json.dumps(payload)
        payload = json.loads(json_str)

        await self._handle_mqtt_message(topic, payload)

    async def _parse_mqtt_response(self, topic: str, payload: dict) -> None:
        """Parse the MQTT response."""
        if topic.endswith("/app/down/thing/events"):
            _LOGGER.debug("Thing event received")
            event = ThingEventMessage.from_dicts(payload)
            params = event.params
            if isinstance(params, dict) or params.identifier is None:
                _LOGGER.debug("Received dict params: %s", params)
                return
            if params.identifier == "device_protobuf_msg_event" and event.method == "thing.events":
                _LOGGER.debug("Protobuf event")
                # Call the callbacks for each cloudDevice
                await self.mqtt_message_event.data_event(event)
            if event.method == "thing.properties":
                _LOGGER.debug(event)

    async def _handle_mqtt_message(self, topic: str, payload: dict) -> None:
        """Async handler for incoming MQTT messages."""
        await self._parse_mqtt_response(topic=topic, payload=payload)

    def _disconnect(self) -> None:
        """Disconnect the MQTT client."""
        self._mqtt_client.disconnect()

    @property
    def waiting_queue(self):
        return self._waiting_queue


class MammotionBaseCloudDevice(MammotionBaseDevice):
    """Base class for Mammotion Cloud devices."""

    def __init__(self, mqtt: MammotionCloud, cloud_device: Device, mowing_state: MowingDevice) -> None:
        """Initialize MammotionBaseCloudDevice."""
        super().__init__(mowing_state, cloud_device)
        self.on_ready_callback: Optional[Callable[[], Awaitable[None]]] = None
        self.loop = asyncio.get_event_loop()
        self._mqtt = mqtt
        self.iot_id = cloud_device.iotId
        self.device = cloud_device
        self._mower = mowing_state
        self._command_futures = {}
        self._commands: MammotionCommand = MammotionCommand(cloud_device.deviceName)
        self.currentID = ""
        self._mqtt.mqtt_message_event.add_subscribers(self._parse_message_for_device)
        self._mqtt.on_ready_event.add_subscribers(self.on_ready)
        self.set_queue_callback(self.queue_command)

        if self._mqtt.is_ready:
            self.run_periodic_sync_task()

    async def on_ready(self) -> None:
        """Callback for when MQTT is subscribed to events."""
        await self._ble_sync()
        await self.run_periodic_sync_task()
        if self.on_ready_callback:
            await self.on_ready_callback()

    async def _ble_sync(self) -> None:
        command_bytes = self._commands.send_todev_ble_sync(3)
        loop = asyncio.get_running_loop()
        await loop.run_in_executor(None, self._mqtt.send_command, self.iot_id, command_bytes)

    async def run_periodic_sync_task(self) -> None:
        """Send ble sync to robot."""
        try:
            if not self._mqtt._operation_lock.locked():
                await self._ble_sync()
        finally:
            self.schedule_ble_sync()

    def schedule_ble_sync(self) -> None:
        """Periodically sync to keep connection alive."""
        if self._mqtt is not None and self._mqtt.is_connected:
            self._ble_sync_task = self.loop.call_later(
                160, lambda: asyncio.ensure_future(self.run_periodic_sync_task())
            )

    async def queue_command(self, key: str, **kwargs: Any) -> bytes:
        # Create a future to hold the result
        _LOGGER.debug("Queueing command: %s", key)
        future = asyncio.Future()
        # Put the command in the queue as a tuple (key, command, future)
        command_bytes = getattr(self._commands, key)(**kwargs)
        await self._mqtt.command_queue.put((self.iot_id, key, command_bytes, future))
        # Wait for the future to be resolved
        return await future

    def _extract_message_id(self, payload: dict) -> str:
        """Extract the message ID from the payload."""
        return payload.get("id", "")

    def _extract_encoded_message(self, payload: dict) -> str:
        """Extract the encoded message from the payload."""
        try:
            content = payload.get("data", {}).get("data", {}).get("params", {}).get("content", "")
            return str(content)
        except AttributeError:
            _LOGGER.error("Error extracting encoded message. Payload: %s", payload)
            return ""

    @staticmethod
    def dequeue_by_iot_id(queue, iot_id):
        for item in queue:
            if item.iot_id == iot_id:
                queue.remove(item)
                return item
        return None

    async def _parse_message_for_device(self, event: ThingEventMessage) -> None:
        params = event.params
        if event.params.iotId != self.iot_id:
            return
        binary_data = base64.b64decode(params.value.content)
        self._update_raw_data(binary_data)
        new_msg = LubaMsg().parse(binary_data)

        if (
            self._commands.get_device_product_key() == ""
            and self._commands.get_device_name() == event.params.deviceName
        ):
            self._commands.set_device_product_key(event.params.productKey)

        if betterproto.serialized_on_wire(new_msg.net):
            if new_msg.net.todev_ble_sync != 0 or has_field(new_msg.net.toapp_wifi_iot_status):
                return

        if len(self._mqtt.waiting_queue) > 0:
            fut: MammotionFuture = self.dequeue_by_iot_id(self._mqtt.waiting_queue, self.iot_id)
            if fut is None:
                return
            while fut.fut.cancelled() and len(self._mqtt.waiting_queue) > 0:
                fut: MammotionFuture = self.dequeue_by_iot_id(self._mqtt.waiting_queue, self.iot_id)
            if not fut.fut.cancelled():
                fut.resolve(cast(bytes, binary_data))
        await self._state_manager.notification(new_msg)

    @property
    def mqtt(self):
        return self._mqtt
